// Copyright (c) 2013 by István Váradi

// This file is part of VSCPL, a simple cross-platform utility library

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// The views and conclusions contained in the software and documentation are those
// of the authors and should not be interpreted as representing official policies,
// either expressed or implied, of the FreeBSD Project.

#ifndef HU_VARADIISTVAN_SCPL_WIN32_CONDVAR_H
#define HU_VARADIISTVAN_SCPL_WIN32_CONDVAR_H
//------------------------------------------------------------------------------

#include "Mutex.h"

//------------------------------------------------------------------------------

namespace hu { namespace varadiistvan { namespace scpl {

//------------------------------------------------------------------------------


/**
 * A condition variable for the Win32 API.
 *
 * Note: the CONDITION_VARIABLE structure and the associated functions
 * are available only starting with the Windows Vista API. However,
 * for the time being we also want to support Windows XP, hence this
 * rather fancy implementation.
 */
class CondVar
{
private:
    /**
     * The mutex protecting the internal data structures.
     */
    Mutex mutex;

    /**
     * The number of waiters.
     */
    size_t numWaiters;

    /**
     * The wait generation.
     */
    size_t waitGeneration;

    /**
     * The number of waiters to release.
     */
    size_t numToRelease;

    /**
     * The event to use for waiting.
     */
    HANDLE event;

public:
    /**
     * Construct the condition variable.
     */
    CondVar();

    /**
     * Destroy the condition variable.
     */
    ~CondVar();

    /**
     * Wait for the condition variable.
     *
     * @param mutex the mutex to unlock and re-lock.
     * @param timeout the number of milliseconds to wait at most. If
     * negative, the function will wait indefinitely.
     *
     * @return true, if the condition variable was signalled, false
     * if a timeout occured
     */
    bool wait(Mutex& inMutex, int timeout = -1);

    /**
     * Signal the condition variable to wake up one thread.
     */
    void signal();

    /**
     * Signal all threads waiting on the condition variable.
     */
    void broadcast();
};

//------------------------------------------------------------------------------
// Inline definitions
//------------------------------------------------------------------------------

inline CondVar::CondVar() :
    numWaiters(0),
    waitGeneration(0),
    numToRelease(0)
{
    event  = CreateEvent(0, true, false, 0);
}

//------------------------------------------------------------------------------

inline CondVar::~CondVar()
{
    CloseHandle(event);
}

//------------------------------------------------------------------------------

inline void CondVar::signal()
{
    mutex.lock();
    if (numWaiters>numToRelease) {
        SetEvent(event);
        ++numToRelease;
        ++waitGeneration;
    }
    mutex.unlock();
}

//------------------------------------------------------------------------------

inline void CondVar::broadcast()
{
    mutex.lock();
    if (numWaiters>0) {
        SetEvent(event);
        numToRelease = numWaiters;
        ++waitGeneration;
    }
    mutex.unlock();
}

//------------------------------------------------------------------------------

} /* namespace hu::varadiistvan::scpl */ } /* namespace hu::varadiistvan */ } /* namespace hu */

//------------------------------------------------------------------------------
#endif // HU_VARADIISTVAN_SCPL_WIN32_CONDVAR_H

// Local Variables:
// mode: C++
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
